# -*- coding: utf-8 -*-
"""Constants for the Mantarray File Manager."""
import uuid

from immutabledict import immutabledict

CURI_BIO_ACCOUNT_UUID = uuid.UUID("73f52be0-368c-42d8-a1fd-660d49ba5604")
CURI_BIO_USER_ACCOUNT_ID = uuid.UUID("455b93eb-c78f-4494-9f73-d3291130f126")

MIN_SUPPORTED_FILE_VERSION = "0.1.1"
CURRENT_HDF5_FILE_FORMAT_VERSION = "0.4.2"
FILE_FORMAT_VERSION_METADATA_KEY = "File Format Version"
FILE_MIGRATION_PATHS = immutabledict(
    {
        "0.3.1": "0.4.1",
        "0.4.1": "0.4.2",
    }
)

NOT_APPLICABLE_H5_METADATA = uuid.UUID(
    "59d92e00-99d5-4460-9a28-5a1a0fe9aecf"
)  # Eli (1/19/21): H5 files can't store the concept of `None` in their metadata, so using this value to denote that a particular piece of metadata is not available (i.e. after migrating to a newer file format version)

HARDWARE_TEST_RECORDING_UUID = uuid.UUID("a2e76058-08cd-475d-a55d-31d401c3cb34")
UTC_BEGINNING_DATA_ACQUISTION_UUID = uuid.UUID("98c67f22-013b-421a-831b-0ea55df4651e")
START_RECORDING_TIME_INDEX_UUID = uuid.UUID("e41422b3-c903-48fd-9856-46ff56a6534c")
UTC_BEGINNING_RECORDING_UUID = uuid.UUID("d2449271-0e84-4b45-a28b-8deab390b7c2")
UTC_FIRST_TISSUE_DATA_POINT_UUID = uuid.UUID("b32fb8cb-ebf8-4378-a2c0-f53a27bc77cc")
UTC_FIRST_REF_DATA_POINT_UUID = uuid.UUID("7cc07b2b-4146-4374-b8f3-1c4d40ff0cf7")
CUSTOMER_ACCOUNT_ID_UUID = uuid.UUID("4927c810-fbf4-406f-a848-eba5308576e6")
USER_ACCOUNT_ID_UUID = uuid.UUID("7282cf00-2b6e-4202-9d9e-db0c73c3a71f")
SOFTWARE_BUILD_NUMBER_UUID = uuid.UUID("b4db8436-10a4-4359-932d-aa80e6de5c76")
SOFTWARE_RELEASE_VERSION_UUID = uuid.UUID("432fc3c1-051b-4604-bc3d-cc0d0bd75368")
MAIN_FIRMWARE_VERSION_UUID = uuid.UUID("faa48a0c-0155-4234-afbf-5e5dbaa59537")
SLEEP_FIRMWARE_VERSION_UUID = uuid.UUID("3a816076-90e4-4437-9929-dc910724a49d")
XEM_SERIAL_NUMBER_UUID = uuid.UUID("e5f5b134-60c7-4881-a531-33aa0edba540")
MANTARRAY_NICKNAME_UUID = uuid.UUID("0cdec9bb-d2b4-4c5b-9dd5-6a49766c5ed4")
MANTARRAY_SERIAL_NUMBER_UUID = uuid.UUID("83720d36-b941-4d85-9b39-1d817799edd6")
REFERENCE_VOLTAGE_UUID = uuid.UUID("0b3f3f56-0cc7-45f0-b748-9b9de480cba8")
WELL_NAME_UUID = uuid.UUID("6d78f3b9-135a-4195-b014-e74dee70387b")
WELL_ROW_UUID = uuid.UUID("da82fe73-16dd-456a-ac05-0b70fb7e0161")
WELL_COLUMN_UUID = uuid.UUID("7af25a0a-8253-4d32-98c4-3c2ca0d83906")
WELL_INDEX_UUID = uuid.UUID("cd89f639-1e36-4a13-a5ed-7fec6205f779")
TOTAL_WELL_COUNT_UUID = uuid.UUID("7ca73e1c-9555-4eca-8281-3f844b5606dc")
REF_SAMPLING_PERIOD_UUID = uuid.UUID("48aa034d-8775-453f-b135-75a983d6b553")
TISSUE_SAMPLING_PERIOD_UUID = uuid.UUID("f629083a-3724-4100-8ece-c03e637ac19c")
ADC_GAIN_SETTING_UUID = uuid.UUID("a3c3bb32-9b92-4da1-8ed8-6c09f9c816f8")
ADC_TISSUE_OFFSET_UUID = uuid.UUID("41069860-159f-49f2-a59d-401783c1ecb4")
ADC_REF_OFFSET_UUID = uuid.UUID("dc10066c-abf2-42b6-9b94-5e52d1ea9bfc")
PLATE_BARCODE_UUID = uuid.UUID("cf60afef-a9f0-4bc3-89e9-c665c6bb6941")
BACKEND_LOG_UUID = uuid.UUID("87533deb-2495-4430-bce7-12fdfc99158e")
COMPUTER_NAME_HASH_UUID = uuid.UUID("fefd0675-35c2-45f6-855a-9500ad3f100d")
BARCODE_IS_FROM_SCANNER_UUID = uuid.UUID("7d026e86-da70-4464-9181-dc0ce2d47bd1")
IS_FILE_ORIGINAL_UNTRIMMED_UUID = uuid.UUID("52231a24-97a3-497a-917c-86c780d9993f")
TRIMMED_TIME_FROM_ORIGINAL_START_UUID = uuid.UUID(
    "371996e6-5e2d-4183-a5cf-06de7058210a"
)
TRIMMED_TIME_FROM_ORIGINAL_END_UUID = uuid.UUID("55f6770d-c369-42ce-a437-5ed89c3cb1f8")
ORIGINAL_FILE_VERSION_UUID = uuid.UUID("cd1b4063-4a87-4a57-bc12-923ff4890844")
UTC_TIMESTAMP_OF_FILE_VERSION_MIGRATION_UUID = uuid.UUID(
    "399b2148-09d4-418b-a132-e37df2721938"
)
FILE_VERSION_PRIOR_TO_MIGRATION_UUID = uuid.UUID("11b4945b-3cf3-4f67-8bee-7abc3c449756")
METADATA_UUID_DESCRIPTIONS = immutabledict(
    {
        HARDWARE_TEST_RECORDING_UUID: "Is Hardware Test Recording",
        START_RECORDING_TIME_INDEX_UUID: "Timepoint of Beginning of Recording",
        UTC_BEGINNING_DATA_ACQUISTION_UUID: "UTC Timestamp of Beginning of Data Acquisition",
        UTC_BEGINNING_RECORDING_UUID: "UTC Timestamp of Beginning of Recording",
        UTC_FIRST_TISSUE_DATA_POINT_UUID: "UTC Timestamp of Beginning of Recorded Tissue Sensor Data",
        UTC_FIRST_REF_DATA_POINT_UUID: "UTC Timestamp of Beginning of Recorded Reference Sensor Data",
        CUSTOMER_ACCOUNT_ID_UUID: "Customer Account ID",
        USER_ACCOUNT_ID_UUID: "User Account ID",
        SOFTWARE_BUILD_NUMBER_UUID: "Software Build Number",
        SOFTWARE_RELEASE_VERSION_UUID: "Software Release Version",
        MAIN_FIRMWARE_VERSION_UUID: "Firmware Version (Main Controller)",
        SLEEP_FIRMWARE_VERSION_UUID: "Firmware Version (Sleep Mode)",
        XEM_SERIAL_NUMBER_UUID: "XEM Serial Number",
        MANTARRAY_NICKNAME_UUID: "Mantarray Nickname",
        MANTARRAY_SERIAL_NUMBER_UUID: "Mantarray Serial Number",
        REFERENCE_VOLTAGE_UUID: "Reference Voltage",
        WELL_NAME_UUID: "Well Name",
        WELL_ROW_UUID: "Well Row (zero-based)",
        WELL_COLUMN_UUID: "Well Column (zero-based)",
        WELL_INDEX_UUID: "Well Index (zero-based)",
        TOTAL_WELL_COUNT_UUID: "Total Wells in Plate",
        REF_SAMPLING_PERIOD_UUID: "Reference Sensor Sampling Period (microseconds)",
        TISSUE_SAMPLING_PERIOD_UUID: "Tissue Sensor Sampling Period (microseconds)",
        ADC_GAIN_SETTING_UUID: "ADC Gain Setting",
        ADC_TISSUE_OFFSET_UUID: "ADC Tissue Sensor Offset",
        ADC_REF_OFFSET_UUID: "ADC Reference Sensor Offset",
        PLATE_BARCODE_UUID: "Plate Barcode",
        BACKEND_LOG_UUID: "Backend log file identifier",
        COMPUTER_NAME_HASH_UUID: "SHA512 digest of computer name",
        BARCODE_IS_FROM_SCANNER_UUID: "Is this barcode obtained from the scanner",
        IS_FILE_ORIGINAL_UNTRIMMED_UUID: "Is this an original file straight from the instrument and untrimmed",
        TRIMMED_TIME_FROM_ORIGINAL_START_UUID: "Number of centimilliseconds that has been trimmed off the beginning of when the original data started",
        TRIMMED_TIME_FROM_ORIGINAL_END_UUID: "Number of centimilliseconds that has been trimmed off the end of when the original data ended",
        ORIGINAL_FILE_VERSION_UUID: "The original version of the file when recorded, prior to any migrations to newer versions/formats.",
        UTC_TIMESTAMP_OF_FILE_VERSION_MIGRATION_UUID: "Timestamp when this file was migrated from an earlier version.",
        FILE_VERSION_PRIOR_TO_MIGRATION_UUID: "File format version that this file was migrated from",
    }
)
DATETIME_STR_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
CENTIMILLISECONDS_PER_SECOND = 1e5
MICROSECONDS_PER_CENTIMILLISECOND = 10
TISSUE_SENSOR_READINGS = "tissue_sensor_readings"
REFERENCE_SENSOR_READINGS = "reference_sensor_readings"



# -*- coding: utf-8 -*-
"""Exceptions."""
from typing import TYPE_CHECKING
from uuid import UUID

from immutable_data_validation import is_uuid

from .constants import CURRENT_HDF5_FILE_FORMAT_VERSION
from .constants import FILE_MIGRATION_PATHS
from .constants import METADATA_UUID_DESCRIPTIONS
from .constants import MIN_SUPPORTED_FILE_VERSION

if TYPE_CHECKING:
    from .files import WellFile


class WellRecordingsNotFromSameSessionError(Exception):
    def __init__(self, main_well_file: "WellFile", new_well_file: "WellFile"):
        super().__init__(
            f"Previously loaded files for this Plate Recording session were from barcode '{main_well_file.get_plate_barcode()}' taken at {main_well_file.get_begin_recording()}. A new file is attempting to be added that is from barcode '{new_well_file.get_plate_barcode()}' taken at {new_well_file.get_begin_recording()}"
        )


class UnsupportedMantarrayFileVersionError(Exception):
    def __init__(self, file_version: str):
        super().__init__(
            f"Mantarray files of version {file_version} are not supported. The minimum supported file version is {MIN_SUPPORTED_FILE_VERSION}"
        )


class FileAttributeNotFoundError(Exception):
    """Error raised if attempting to access a non-existent attribute."""

    def __init__(
        self,
        attr_name: str,
        file_version: str,
        file_path: str,
    ):
        try:
            attr_description = (
                f"{attr_name}, ({METADATA_UUID_DESCRIPTIONS[UUID(attr_name)]})"
                if is_uuid(attr_name)
                else f"{attr_name} (no UUID given)"
            )
        except KeyError:
            attr_description = f"{attr_name} (Unrecognized UUID)"
        super().__init__(
            f"The metadata attribute '{attr_description}' was not found in this file. File format version {file_version}, filepath: {file_path}"
        )


class UnsupportedFileMigrationPath(Exception):
    """Error raised if a file is attempted to be migrated from a version that has no migration script."""

    def __init__(self, file_version: str):
        super().__init__(
            f"There is no supported migration path from version {file_version}. Supported paths are: {FILE_MIGRATION_PATHS}."
        )


class UnsupportedArgumentError(Exception):
    """Error raised if the arguments indicating the amount of centimilliseconds to be trimmed from the start and end are both None."""

    def __init__(self) -> None:
        super().__init__("Both arguments cannot be None or 0.")


class TooTrimmedError(Exception):
    def __init__(self, from_start: int, from_end: int, total_time: int) -> None:
        super().__init__(
            f"When trimming {from_start} centimilliseconds from the start and {from_end} centimilliseconds from the end, the length of the recording is exceeded. The length of the recording is {total_time} centimilliseconds"
        )


class MantarrayFileNotLatestVersionError(Exception):
    def __init__(self, file_version: str):
        super().__init__(
            f"Mantarray files of version {file_version} are not supported. Please migrate to the latest file version {CURRENT_HDF5_FILE_FORMAT_VERSION}"
        )



# -*- coding: utf-8 -*-
"""Classes and functions for finding and reading files."""
import datetime
from glob import glob
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Union
from uuid import UUID

import h5py
from immutabledict import immutabledict
from nptyping import NDArray
import numpy as np
from semver import VersionInfo
from stdlib_utils import get_current_file_abs_directory

from .constants import CUSTOMER_ACCOUNT_ID_UUID
from .constants import DATETIME_STR_FORMAT
from .constants import FILE_FORMAT_VERSION_METADATA_KEY
from .constants import IS_FILE_ORIGINAL_UNTRIMMED_UUID
from .constants import MANTARRAY_SERIAL_NUMBER_UUID
from .constants import MICROSECONDS_PER_CENTIMILLISECOND
from .constants import MIN_SUPPORTED_FILE_VERSION
from .constants import PLATE_BARCODE_UUID
from .constants import REF_SAMPLING_PERIOD_UUID
from .constants import REFERENCE_SENSOR_READINGS
from .constants import START_RECORDING_TIME_INDEX_UUID
from .constants import TISSUE_SAMPLING_PERIOD_UUID
from .constants import TISSUE_SENSOR_READINGS
from .constants import TRIMMED_TIME_FROM_ORIGINAL_START_UUID
from .constants import USER_ACCOUNT_ID_UUID
from .constants import UTC_BEGINNING_DATA_ACQUISTION_UUID
from .constants import UTC_BEGINNING_RECORDING_UUID
from .constants import UTC_FIRST_REF_DATA_POINT_UUID
from .constants import UTC_FIRST_TISSUE_DATA_POINT_UUID
from .constants import WELL_INDEX_UUID
from .constants import WELL_NAME_UUID
from .exceptions import FileAttributeNotFoundError
from .exceptions import UnsupportedMantarrayFileVersionError
from .exceptions import WellRecordingsNotFromSameSessionError

PATH_OF_CURRENT_FILE = get_current_file_abs_directory()


def _get_file_attr(h5_file: h5py.File, attr_name: str, file_version: str) -> Any:
    if attr_name not in h5_file.attrs:
        file_path = h5_file.filename
        raise FileAttributeNotFoundError(attr_name, file_version, file_path)
    return h5_file.attrs[attr_name]


def get_unique_files_from_directory(directory: str) -> List[str]:
    """Obtain a list of all unique h5 files in the current directory.

    Args:
        directory: the master folder for which all h5 files reside

    Returns:
        A list of the file paths for all the h5 files in the directory.
    """
    unique_files: List[str] = []

    for path, _, files in os.walk(directory):
        for name in files:
            if not name.endswith(".h5"):
                continue

            file = os.path.join(path, name)
            well = WellFile(file)
            index = well.get_well_index()
            barcode = well.get_plate_barcode()
            start_time = well.get_begin_recording()

            for item in unique_files:
                new_well = WellFile(item)
                if (
                    new_well.get_well_index() == index
                    and new_well.get_plate_barcode() == barcode
                    and new_well.get_begin_recording() == start_time
                ):
                    break
            else:
                unique_files.append(file)

    return unique_files


def get_specified_files(
    search_criteria: str, criteria_value: str, unique_files: List[str]
) -> Dict[str, Dict[str, List[str]]]:
    """Obtain a subset of all the h5 files based on search criteria from user.

    Args:
        search_criteria: A str representing a UUID to a piece a metadata to filter results

    Returns:
        a dictionary of the Plate Recordings that fall under the specified search criteria.
    """
    value_dict: Dict[str, List[str]] = {}
    full_dict: Dict[str, Dict[str, List[str]]] = {}
    plate_recording_list: List[str] = []

    for file in unique_files:
        well = WellFile(file)
        if search_criteria == "Well Name" and well.get_well_name() == criteria_value:
            plate_recording_list.append(file)
        if (
            search_criteria == "Plate Barcode"
            and well.get_plate_barcode() == criteria_value
        ):
            plate_recording_list.append(file)
        if search_criteria == "User ID" and well.get_user_account() == criteria_value:
            plate_recording_list.append(file)
        if (
            search_criteria == "Account ID"
            and well.get_customer_account() == criteria_value
        ):
            plate_recording_list.append(file)
        if (
            search_criteria == "Mantarray Serial Number"
            and well.get_mantarray_serial_number() == criteria_value
        ):
            plate_recording_list.append(file)

    value_dict = {criteria_value: plate_recording_list}
    full_dict = {search_criteria: value_dict}

    return full_dict


def _extract_datetime_from_h5(
    open_h5_file: h5py.File,
    file_version: str,
    metadata_uuid: UUID,
) -> datetime.datetime:
    if file_version.split(".") < VersionInfo.parse("0.2.1"):
        if metadata_uuid == UTC_BEGINNING_RECORDING_UUID:
            # Tanner (9/17/20): The use of this proxy value is justified by the fact that there is a 15 second delay between when data is recorded and when the GUI displays it, and because the GUI will send the timestamp of when the recording button is pressed.
            acquisition_timestamp_str = _get_file_attr(
                open_h5_file,
                str(UTC_BEGINNING_DATA_ACQUISTION_UUID),
                file_version,
            )
            begin_recording = datetime.datetime.strptime(
                acquisition_timestamp_str, DATETIME_STR_FORMAT
            ).replace(tzinfo=datetime.timezone.utc) + datetime.timedelta(seconds=15)
            return begin_recording
        if metadata_uuid == UTC_FIRST_TISSUE_DATA_POINT_UUID:
            # Tanner (9/17/20): Early file versions did not include this metadata under a UUID, so we have to use this string identifier instead
            metadata_name = "UTC Timestamp of Beginning of Recorded Tissue Sensor Data"
            timestamp_str = _get_file_attr(
                open_h5_file,
                str(metadata_name),
                file_version,
            )
            return datetime.datetime.strptime(
                timestamp_str, DATETIME_STR_FORMAT
            ).replace(tzinfo=datetime.timezone.utc)
        if metadata_uuid == UTC_FIRST_REF_DATA_POINT_UUID:
            # Tanner (10/5/20): Early file versions did not include this metadata under a UUID, so we have to use this string identifier instead
            metadata_name = (
                "UTC Timestamp of Beginning of Recorded Reference Sensor Data"
            )
            timestamp_str = _get_file_attr(
                open_h5_file,
                str(metadata_name),
                file_version,
            )
            return datetime.datetime.strptime(
                timestamp_str, DATETIME_STR_FORMAT
            ).replace(tzinfo=datetime.timezone.utc)

    timestamp_str = _get_file_attr(open_h5_file, str(metadata_uuid), file_version)
    return datetime.datetime.strptime(timestamp_str, DATETIME_STR_FORMAT).replace(
        tzinfo=datetime.timezone.utc
    )


class BasicWellFile:
    """Very thin wrapper around an H5 file for a single well of data.

    Used typically just for assessing file version when migrating.

    Args:
        file_name: The path of the H5 file to open.

    Attributes:
        _h5_file: The opened H5 file object.
    """

    def __init__(self, file_name: str) -> None:
        self._h5_file: h5py.File = h5py.File(
            file_name,
            "r",
        )
        self._file_name = file_name
        self._file_version: str = self._h5_file.attrs[FILE_FORMAT_VERSION_METADATA_KEY]

    def get_h5_file(self) -> h5py.File:
        return self._h5_file

    def get_file_name(self) -> str:
        return self._file_name

    def get_file_version(self) -> str:
        return self._file_version

    def get_h5_attribute(self, attr_name: str) -> Any:
        return _get_file_attr(self._h5_file, attr_name, self._file_version)

    def __del__(self) -> None:
        self._h5_file.close()


class WellFileMixIn:
    # pylint: disable=too-few-public-methods # Eli (1/18/21): If these mixins don't have these base attributes then mypy complains
    _file_version: str
    _h5_file: h5py.File


class PlateMetadataMixIn(WellFileMixIn):
    # pylint: disable=too-few-public-methods # Eli (1/18/21): I think its better to keep these MixIns well separated for better future flexibility
    def get_plate_barcode(self) -> str:
        return str(
            _get_file_attr(
                self._h5_file,
                str(PLATE_BARCODE_UUID),
                self._file_version,
            )
        )


class WellMetadataMixIn(WellFileMixIn):
    """Mixin for metadata related to the well itself."""

    def get_well_name(self) -> str:
        return str(
            _get_file_attr(
                self._h5_file,
                str(WELL_NAME_UUID),
                self._file_version,
            )
        )

    def get_well_index(self) -> int:
        return int(
            _get_file_attr(
                self._h5_file,
                str(WELL_INDEX_UUID),
                self._file_version,
            )
        )


class InstrumentMetadataMixIn(WellFileMixIn):
    # pylint: disable=too-few-public-methods # Eli (1/18/21): I think its better to keep these MixIns well separated for better future flexibility
    def get_mantarray_serial_number(self) -> str:
        return str(
            _get_file_attr(
                self._h5_file,
                str(MANTARRAY_SERIAL_NUMBER_UUID),
                self._file_version,
            )
        )


class CustomerMetadataMixIn(WellFileMixIn):
    """Mixin for metadata related to the customer."""

    def get_user_account(self) -> UUID:
        return UUID(
            _get_file_attr(
                self._h5_file,
                str(USER_ACCOUNT_ID_UUID),
                self._file_version,
            )
        )

    def get_customer_account(self) -> UUID:
        return UUID(
            _get_file_attr(
                self._h5_file,
                str(CUSTOMER_ACCOUNT_ID_UUID),
                self._file_version,
            )
        )


class LikelyConsistentMetadata(
    CustomerMetadataMixIn,
    InstrumentMetadataMixIn,
    PlateMetadataMixIn,
    WellMetadataMixIn,
):
    """A wrapper for mixins that will likely be in all future file formats."""


class WellFile(
    BasicWellFile, LikelyConsistentMetadata
):  # pylint: disable=too-many-ancestors # Eli (7/28/20): I don't see a way around this...we need to subclass h5py File
    """Wrapper around an H5 file for a single well of data.

    This is only guaranteed to function correctly on the current working file format version.
    Use the file migrate_to_latest_version to get files up to date with the current working version.

    Args:
        file_name: The path of the H5 file to open.

    Attributes:
        _h5_file: The opened H5 file object.
    """

    def __init__(self, file_name: str) -> None:
        super().__init__(file_name)
        self._raw_tissue_reading: Optional[NDArray[(2, Any), int]] = None
        self._raw_ref_reading: Optional[NDArray[(2, Any), int]] = None

    def get_unique_recording_key(self) -> Tuple[str, datetime.datetime]:
        barcode = self.get_plate_barcode()
        start_time = self.get_begin_recording()
        return barcode, start_time

    def get_timestamp_of_beginning_of_data_acquisition(self) -> datetime.datetime:
        return _extract_datetime_from_h5(
            self._h5_file, self._file_version, UTC_BEGINNING_DATA_ACQUISTION_UUID
        )

    def get_begin_recording(self) -> datetime.datetime:
        return _extract_datetime_from_h5(
            self._h5_file, self._file_version, UTC_BEGINNING_RECORDING_UUID
        )

    def get_timestamp_of_first_tissue_data_point(self) -> datetime.datetime:
        return _extract_datetime_from_h5(
            self._h5_file, self._file_version, UTC_FIRST_TISSUE_DATA_POINT_UUID
        )

    def get_timestamp_of_first_ref_data_point(self) -> datetime.datetime:
        return _extract_datetime_from_h5(
            self._h5_file, self._file_version, UTC_FIRST_REF_DATA_POINT_UUID
        )

    def get_tissue_sampling_period_microseconds(self) -> int:
        return int(
            _get_file_attr(
                self._h5_file,
                str(TISSUE_SAMPLING_PERIOD_UUID),
                self._file_version,
            )
        )

    def get_reference_sampling_period_microseconds(self) -> int:
        return int(
            _get_file_attr(
                self._h5_file,
                str(REF_SAMPLING_PERIOD_UUID),
                self._file_version,
            )
        )

    def get_recording_start_index(self) -> int:
        """Get the time index when recording was requested.

        This is the number of centimilliseconds from beginning of the
        start of data acquisition that was displayed on the screen when
        the user pressed the Record button.
        """
        return int(
            _get_file_attr(
                self._h5_file,
                str(START_RECORDING_TIME_INDEX_UUID),
                self._file_version,
            )
        )

    def get_raw_tissue_reading(self) -> NDArray[(2, Any), int]:
        """Get a value vs time array.

        Time (centi-milliseconds) is first dimension, value is second
        dimension.

        Time is given relative to the start of the recording, so that arrays from different wells can be displayed together
        """
        if self._raw_tissue_reading is None:
            recording_start_index_useconds = (
                self.get_recording_start_index() * MICROSECONDS_PER_CENTIMILLISECOND
            )
            timestamp_of_start_index = (
                self.get_timestamp_of_beginning_of_data_acquisition()
                + datetime.timedelta(microseconds=recording_start_index_useconds)
            )
            time_delta = (
                self.get_timestamp_of_first_tissue_data_point()
                - timestamp_of_start_index
            )
            time_delta_centimilliseconds = int(
                time_delta
                / datetime.timedelta(microseconds=MICROSECONDS_PER_CENTIMILLISECOND)
            )

            time_step = int(
                self.get_tissue_sampling_period_microseconds()
                / MICROSECONDS_PER_CENTIMILLISECOND
            )
            tissue_data = self._h5_file[TISSUE_SENSOR_READINGS]

            times = np.arange(len(tissue_data), dtype=np.int32) * time_step
            len_time = len(times)

            time_delta_centimilliseconds = self._check_for_trimmed_file(
                times, time_delta_centimilliseconds
            )

            self._raw_tissue_reading = np.array(
                (times + time_delta_centimilliseconds, tissue_data[:len_time]),
                dtype=np.int32,
            )
        return self._raw_tissue_reading

    def get_raw_reference_reading(self) -> NDArray[(2, Any), int]:
        """Get a reference value vs time array.

        Time (centi-milliseconds) is first dimension, reference value is second
        dimension.

        Time is given relative to the start of the recording, so that arrays from different wells can be displayed together
        """
        if self._raw_ref_reading is None:
            recording_start_index_useconds = (
                self.get_recording_start_index() * MICROSECONDS_PER_CENTIMILLISECOND
            )
            timestamp_of_start_index = (
                self.get_timestamp_of_beginning_of_data_acquisition()
                + datetime.timedelta(microseconds=recording_start_index_useconds)
            )
            time_delta = (
                self.get_timestamp_of_first_ref_data_point() - timestamp_of_start_index
            )

            time_delta_centimilliseconds = int(
                time_delta
                / datetime.timedelta(microseconds=MICROSECONDS_PER_CENTIMILLISECOND)
            )

            time_step = int(
                self.get_reference_sampling_period_microseconds()
                / MICROSECONDS_PER_CENTIMILLISECOND
            )

            ref_data = self._h5_file[REFERENCE_SENSOR_READINGS]

            times = np.arange(len(ref_data), dtype=np.int32) * time_step
            len_time = len(times)

            time_delta_centimilliseconds = self._check_for_trimmed_file(
                times, time_delta_centimilliseconds
            )

            self._raw_ref_reading = np.array(
                (times + time_delta_centimilliseconds, ref_data[:len_time]),
                dtype=np.int32,
            )

        return self._raw_ref_reading

    def _check_for_trimmed_file(
        self, times: NDArray[(1, Any), int], time_delta_centimilliseconds: int
    ) -> int:
        try:
            is_untrimmed = self.get_h5_attribute(str(IS_FILE_ORIGINAL_UNTRIMMED_UUID))
        except FileAttributeNotFoundError:
            # Eli (2/5/21): If that attribute is missing, then it's an old version of the file from before that metadata was added. So therefore it must be an original file that was never trimmed.
            is_untrimmed = True
        if is_untrimmed:
            return time_delta_centimilliseconds
        time_trimmed = self.get_h5_attribute(str(TRIMMED_TIME_FROM_ORIGINAL_START_UUID))
        new_times = times + time_delta_centimilliseconds
        start_index = find_start_index(time_trimmed, new_times)
        new_time_delta = int(new_times[start_index])
        return new_time_delta


def find_start_index(from_start: int, old_data: NDArray[(1, Any), int]) -> int:
    start_index = 0
    time_elapsed = 0
    while start_index + 1 < len(old_data) and from_start >= time_elapsed:
        time_elapsed += old_data[start_index + 1] - old_data[start_index]
        start_index = start_index + 1
    start_index = start_index - 1
    return start_index


class WellFile_0_3_1(  # pylint:disable=invalid-name,too-many-ancestors # Eli (1/18/21): this seems like a good way to specifically name these historical class objects. I don't see a way around this ancestor issue...we need to subclass h5py File
    WellFile
):
    """Historical class to open WellFiles in version 0.3.1.

    Typically kept around for test cases and migration scripts.

    If the main WellFile object stops being compatible, then old
    deprecated methods should be moved here.
    """


class WellFile_0_4_1(  # pylint:disable=invalid-name,too-many-ancestors # Eli (1/18/21): this seems like a good way to specifically name these historical class objects. I don't see a way around this ancestor issue...we need to subclass h5py File
    WellFile
):
    """Historical class to open WellFiles in version 0.4.1.

    Typically kept around for test cases and migration scripts.

    If the main WellFile object stops being compatible, then old
    deprecated methods should be moved here.
    """


class WellFile_0_4_2(  # pylint:disable=invalid-name,too-many-ancestors # Eli (1/18/21): this seems like a good way to specifically name these historical class objects. I don't see a way around this ancestor issue...we need to subclass h5py File
    WellFile
):
    """Historical class to open WellFiles in version 0.4.2.

    Typically kept around for test cases and migration scripts.

    If the main WellFile object stops being compatible, then old
    deprecated methods should be moved here.
    """


class PlateRecording:
    """Wrapper around 24 WellFiles for a single plate of data.

    Args:
        file_paths: A list of all the file paths for each h5 file to open, or already instantiated WellFile objects.

    Attributes:
        _files : WellFiles of all the file paths provided.
    """

    def __init__(self, file_paths: Sequence[Union[str, WellFile]]) -> None:
        self._files: List[WellFile] = list()
        self._wells_by_index: Dict[int, WellFile] = dict()
        min_supported_version = VersionInfo.parse(MIN_SUPPORTED_FILE_VERSION)
        for iter_file_path in file_paths:

            well_file = iter_file_path
            if isinstance(well_file, str):
                well_file = WellFile(well_file)
            file_version_str = well_file.get_file_version()
            if file_version_str.split(".") < min_supported_version:
                raise UnsupportedMantarrayFileVersionError(file_version_str)
            if len(self._files) > 0:
                new_session_key = well_file.get_unique_recording_key()
                old_file = self._files[0]
                old_session_key = old_file.get_unique_recording_key()
                if new_session_key != old_session_key:
                    raise WellRecordingsNotFromSameSessionError(old_file, well_file)
            self._files.append(well_file)
            self._wells_by_index[well_file.get_well_index()] = well_file

    @classmethod
    def from_directory(cls, dir_to_load_files_from: str) -> "PlateRecording":
        return cls(glob(os.path.join(dir_to_load_files_from, "*.h5")))

    def get_well_by_index(self, well_index: int) -> WellFile:
        return self._wells_by_index[well_index]

    def get_wellfile_names(self) -> Sequence[str]:
        well_files = []
        for well in self._files:
            well_files.append(well.get_well_name())
        return well_files

    def get_well_names(self) -> Set[str]:
        out_set: Set[str] = set()
        for iter_well_file in self._files:
            out_set.add(iter_well_file.get_well_name())
        return out_set

    def get_well_indices(self) -> Tuple[int, ...]:
        return tuple(sorted(self._wells_by_index.keys()))


WELL_FILE_CLASSES = immutabledict(
    {"0.3.1": WellFile_0_3_1, "0.4.1": WellFile_0_4_1, "0.4.2": WellFile_0_4_2}
)



# -*- coding: utf-8 -*-
"""Classes and functions for writing and migrating files."""
import datetime
import ntpath
import os
from os import getcwd
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union
import uuid

import h5py
from immutable_data_validation import validate_int
from nptyping import NDArray
import numpy as np

from .constants import BACKEND_LOG_UUID
from .constants import BARCODE_IS_FROM_SCANNER_UUID
from .constants import COMPUTER_NAME_HASH_UUID
from .constants import CURRENT_HDF5_FILE_FORMAT_VERSION
from .constants import DATETIME_STR_FORMAT
from .constants import FILE_FORMAT_VERSION_METADATA_KEY
from .constants import FILE_MIGRATION_PATHS
from .constants import FILE_VERSION_PRIOR_TO_MIGRATION_UUID
from .constants import IS_FILE_ORIGINAL_UNTRIMMED_UUID
from .constants import NOT_APPLICABLE_H5_METADATA
from .constants import ORIGINAL_FILE_VERSION_UUID
from .constants import REFERENCE_SENSOR_READINGS
from .constants import TISSUE_SENSOR_READINGS
from .constants import TRIMMED_TIME_FROM_ORIGINAL_END_UUID
from .constants import TRIMMED_TIME_FROM_ORIGINAL_START_UUID
from .constants import UTC_TIMESTAMP_OF_FILE_VERSION_MIGRATION_UUID
from .exceptions import MantarrayFileNotLatestVersionError
from .exceptions import TooTrimmedError
from .exceptions import UnsupportedArgumentError
from .exceptions import UnsupportedFileMigrationPath
from .files import BasicWellFile
from .files import find_start_index
from .files import WELL_FILE_CLASSES
from .files import WellFile


class MantarrayH5FileCreator(
    h5py.File
):  # pylint: disable=too-many-ancestors # Eli (7/28/20): I don't see a way around this...we need to subclass h5py File
    """Creates an H5 file with the basic format/layout."""

    def __init__(
        self,
        file_name: str,
        file_format_version: str = CURRENT_HDF5_FILE_FORMAT_VERSION,
    ) -> None:
        super().__init__(
            file_name,
            "w",
            libver="latest",  # Eli (2/9/20) tried to specify this ('earliest', 'v110') to be more backward compatible but it didn't work for unknown reasons (gave error when trying to set swmr_mode=True)
            userblock_size=512,  # minimum size is 512 bytes
        )

        self.attrs[FILE_FORMAT_VERSION_METADATA_KEY] = file_format_version


def _get_format_version_of_file(file_path: str) -> str:
    file = BasicWellFile(file_path)
    file_version = file.get_file_version()
    file.get_h5_file().close()
    return file_version


def migrate_to_next_version(
    starting_file_path: str, working_directory: Optional[str] = None
) -> str:
    """Migrates an H5 file to the next version along the migration path.

    Args:
        starting_file_path: the path to the H5 file
        working_directory: the directory in which to create the new files. Defaults to current working directory

    Returns:
        The path to the H5 file migrated to the next version.
    """
    file_version = _get_format_version_of_file(starting_file_path)
    if file_version == CURRENT_HDF5_FILE_FORMAT_VERSION:
        return starting_file_path
    if working_directory is None:
        working_directory = getcwd()
    if file_version not in FILE_MIGRATION_PATHS:
        raise UnsupportedFileMigrationPath(file_version)
    old_file = WELL_FILE_CLASSES[file_version](starting_file_path)
    new_file_version = FILE_MIGRATION_PATHS[file_version]
    old_file_basename = ntpath.basename(starting_file_path)
    old_file_basename_no_suffix = old_file_basename[:-3]

    new_file_name = os.path.join(
        working_directory, f"{old_file_basename_no_suffix}__v{new_file_version}.h5"
    )
    new_file = MantarrayH5FileCreator(
        new_file_name, file_format_version=new_file_version
    )

    # old metadata
    old_h5_file = old_file.get_h5_file()
    old_metadata_keys = set(old_h5_file.attrs.keys())
    old_metadata_keys.remove(FILE_FORMAT_VERSION_METADATA_KEY)
    for iter_metadata_key in old_metadata_keys:
        new_file.attrs[iter_metadata_key] = old_h5_file.attrs[iter_metadata_key]

    # transfer data
    old_tissue_data = old_h5_file["tissue_sensor_readings"]
    new_file.create_dataset("tissue_sensor_readings", data=old_tissue_data)
    old_reference_data = old_h5_file["reference_sensor_readings"]
    new_file.create_dataset("reference_sensor_readings", data=old_reference_data)

    # new metadata
    metadata_to_create: Tuple[Tuple[uuid.UUID, Union[str, bool, int, float]], ...]
    if new_file_version == "0.4.1":
        metadata_to_create = (
            (BARCODE_IS_FROM_SCANNER_UUID, False),
            (IS_FILE_ORIGINAL_UNTRIMMED_UUID, True),
            (TRIMMED_TIME_FROM_ORIGINAL_START_UUID, 0),
            (TRIMMED_TIME_FROM_ORIGINAL_END_UUID, 0),
            (BACKEND_LOG_UUID, str(NOT_APPLICABLE_H5_METADATA)),
            (COMPUTER_NAME_HASH_UUID, str(NOT_APPLICABLE_H5_METADATA)),
        )
    elif new_file_version == "0.4.2":
        utc_now = datetime.datetime.utcnow()
        formatted_time = utc_now.strftime(DATETIME_STR_FORMAT)
        metadata_to_create = (
            (
                ORIGINAL_FILE_VERSION_UUID,
                str(NOT_APPLICABLE_H5_METADATA),
            ),  # Eli (1/19/21): there's no way I can think of to know for sure what the very original file version was since it wasn't recorded as metadata, so just leaving it blank for now.
            (FILE_VERSION_PRIOR_TO_MIGRATION_UUID, file_version),
            (UTC_TIMESTAMP_OF_FILE_VERSION_MIGRATION_UUID, formatted_time),
        )
    else:
        raise NotImplementedError(
            f"Migrating to the version {new_file_version} is not supported."
        )
    for iter_metadata_key, iter_metadata_value in metadata_to_create:
        new_file.attrs[str(iter_metadata_key)] = iter_metadata_value

    new_file.close()
    return new_file_name


def migrate_to_latest_version(
    starting_file_path: str, working_directory: Optional[str] = None
) -> str:
    """Migrates an H5 file to the latest version.

    To use from the command line: `python -c "from mantarray_file_manager import migrate_to_latest_version; migrate_to_latest_version('tests/h5/v0.3.1/MA20123456__2020_08_17_145752__A1.h5')"`

    Args:
        starting_file_path: the path to the H5 file
        working_directory: the directory in which to create the new files. Defaults to current working directory

    Returns:
        The path to the final H5 file migrated to the latest version.
    """
    current_file_path = starting_file_path
    while True:
        file_version = _get_format_version_of_file(current_file_path)
        if file_version == CURRENT_HDF5_FILE_FORMAT_VERSION:
            return current_file_path
        current_file_path = migrate_to_next_version(
            current_file_path, working_directory=working_directory
        )


def h5_file_trimmer(
    file_path: str,
    working_directory: Optional[str] = None,
    from_start: Optional[int] = 0,
    from_end: Optional[int] = 0,
) -> str:
    """Trims an H5 file.

    To use from the command line: `python -c "from mantarray_file_manager import h5_file_trimmer; h5_file_trimmer('tests/h5/v0.4.2/MA190190000__2021_01_19_011931__C3__v0.4.2.h5')"`

    Args:
        file_path: path to the H5 file
        working_directory: the directory in which to create the new files. Defaults to current working directory.
        from_start: centimilliseconds to trim from the start
        from_end: centimilliseconds to trim from the end

    Returns:
        The path to the trimmed H5 file. The amount actually trimmed off the file is dependent on the timepoints of the tissue sensor data and will be reflected in the new file name, message to the terminal, and the metadata. If the amount to be trimmed off is in between two time points, less time will be trimmed off and the lower timepoint will be used if from_start or upper timepoint if from_last. Reference sensor readings are trimmed according to the amount trimmed from tissue data.
    """
    # pylint: disable-msg=too-many-locals # Anna (1/27/20) many local variables are needed throughout the method
    validate_int(
        value=from_start,
        allow_null=True,
        minimum=0,
    )
    validate_int(value=from_end, allow_null=True, minimum=0)

    if from_start == 0 and from_end == 0:
        raise UnsupportedArgumentError()

    if from_end is None or from_start is None:
        raise UnsupportedArgumentError()

    old_file_version = _get_format_version_of_file(file_path)

    if old_file_version != CURRENT_HDF5_FILE_FORMAT_VERSION:
        raise MantarrayFileNotLatestVersionError(old_file_version)

    if working_directory is None:
        working_directory = getcwd()

    old_file = WellFile(file_path)
    old_file_basename = ntpath.basename(file_path)[:-3]

    # finding amount to trim
    old_raw_reference_data = old_file.get_raw_reference_reading()
    old_tissue_data = old_file.get_raw_tissue_reading()

    tissue_data_start_val = old_tissue_data[0][0]
    tissue_data_last_val = old_tissue_data[0][-1]
    total_time = tissue_data_last_val - tissue_data_start_val
    tissue_data_start_index = find_start_index(from_start, old_tissue_data[0])
    tissue_data_last_index = _find_last_index(from_end, old_tissue_data)

    actual_start_trimmed = (
        old_tissue_data[0][tissue_data_start_index] - tissue_data_start_val
    )
    actual_end_trimmed = (
        tissue_data_last_val - old_tissue_data[0][tissue_data_last_index]
    )

    reference_data_start_index = find_start_index(
        actual_start_trimmed, old_raw_reference_data[0]
    )
    reference_data_last_index = _find_last_index(
        actual_end_trimmed, old_raw_reference_data
    )

    if (
        reference_data_start_index >= reference_data_last_index
        or tissue_data_start_index >= tissue_data_last_index
    ):
        raise TooTrimmedError(from_start, from_end, total_time)

    # old metadata
    old_h5_file = old_file.get_h5_file()
    old_metadata_keys = set(old_h5_file.attrs.keys())
    old_from_end = 0
    old_from_start = 0
    is_untrimmed = old_h5_file.attrs[str(IS_FILE_ORIGINAL_UNTRIMMED_UUID)]

    if not is_untrimmed:
        old_from_start = old_h5_file.attrs[str(TRIMMED_TIME_FROM_ORIGINAL_START_UUID)]
        old_from_end = old_h5_file.attrs[str(TRIMMED_TIME_FROM_ORIGINAL_END_UUID)]

        old_metadata_keys.remove(str(TRIMMED_TIME_FROM_ORIGINAL_START_UUID))
        old_metadata_keys.remove(str(TRIMMED_TIME_FROM_ORIGINAL_END_UUID))
        old_metadata_keys.remove(str(IS_FILE_ORIGINAL_UNTRIMMED_UUID))

        old_file_basename = old_file_basename.split("__trimmed")[0]

    if actual_start_trimmed != from_start:

        print(  # allow-print
            f"{actual_start_trimmed} centimilliseconds was trimmed from the start instead of {from_start}"
        )

    if actual_end_trimmed != from_end:

        print(  # allow-print
            f"{actual_end_trimmed} centimilliseconds was trimmed from the end instead of {from_end}"
        )

    # create new file
    new_file_name = os.path.join(
        working_directory,
        f"{old_file_basename}__trimmed_{actual_start_trimmed + old_from_start}_{actual_end_trimmed + old_from_end}.h5",
    )

    new_file = MantarrayH5FileCreator(new_file_name)

    for iter_metadata_key in old_metadata_keys:
        new_file.attrs[iter_metadata_key] = old_h5_file.attrs[iter_metadata_key]

    # new metadata
    metadata_to_create: Tuple[Tuple[uuid.UUID, Union[str, bool, int, float]], ...]

    metadata_to_create = (
        (IS_FILE_ORIGINAL_UNTRIMMED_UUID, False),
        (TRIMMED_TIME_FROM_ORIGINAL_START_UUID, actual_start_trimmed + old_from_start),
        (TRIMMED_TIME_FROM_ORIGINAL_END_UUID, actual_end_trimmed + old_from_end),
    )

    for iter_metadata_key, iter_metadata_value in metadata_to_create:
        new_file.attrs[str(iter_metadata_key)] = iter_metadata_value

    # adding new trimmed data
    new_tissue_sensor_data = old_h5_file[TISSUE_SENSOR_READINGS]
    new_tissue_sensor_data = np.array(new_tissue_sensor_data)
    new_tissue_sensor_data = new_tissue_sensor_data[
        tissue_data_start_index : tissue_data_last_index + 1
    ]  # +1 because needs to be inclusive of last index

    new_reference_sensor_data = old_h5_file[REFERENCE_SENSOR_READINGS]
    new_reference_sensor_data = np.array(new_reference_sensor_data)
    new_reference_sensor_data = new_reference_sensor_data[
        reference_data_start_index : reference_data_last_index + 1
    ]  # +1 because needs to be indclusive of last index

    new_file.create_dataset(TISSUE_SENSOR_READINGS, data=new_tissue_sensor_data)
    new_file.create_dataset(REFERENCE_SENSOR_READINGS, data=new_reference_sensor_data)

    old_h5_file.close()
    new_file.close()
    return new_file_name


def _find_last_index(from_end: int, old_data: NDArray[(2, Any), int]) -> int:
    last_index = len(old_data[0]) - 1
    time_elapsed = 0
    while last_index > 0 and from_end >= time_elapsed:
        time_elapsed += old_data[0][last_index] - old_data[0][last_index - 1]
        last_index -= 1
    last_index += 1
    return last_index



# -*- coding: utf-8 -*-
"""Curi Bio File Manager.

File Manager for utilizing Curi bio data files and online databases.
"""
from . import file_writer
from .constants import ADC_GAIN_SETTING_UUID
from .constants import ADC_REF_OFFSET_UUID
from .constants import ADC_TISSUE_OFFSET_UUID
from .constants import BACKEND_LOG_UUID
from .constants import BARCODE_IS_FROM_SCANNER_UUID
from .constants import CENTIMILLISECONDS_PER_SECOND
from .constants import COMPUTER_NAME_HASH_UUID
from .constants import CURI_BIO_ACCOUNT_UUID
from .constants import CURI_BIO_USER_ACCOUNT_ID
from .constants import CURRENT_HDF5_FILE_FORMAT_VERSION
from .constants import CUSTOMER_ACCOUNT_ID_UUID
from .constants import DATETIME_STR_FORMAT
from .constants import FILE_FORMAT_VERSION_METADATA_KEY
from .constants import FILE_MIGRATION_PATHS
from .constants import FILE_VERSION_PRIOR_TO_MIGRATION_UUID
from .constants import HARDWARE_TEST_RECORDING_UUID
from .constants import IS_FILE_ORIGINAL_UNTRIMMED_UUID
from .constants import MAIN_FIRMWARE_VERSION_UUID
from .constants import MANTARRAY_NICKNAME_UUID
from .constants import MANTARRAY_SERIAL_NUMBER_UUID
from .constants import METADATA_UUID_DESCRIPTIONS
from .constants import MICROSECONDS_PER_CENTIMILLISECOND
from .constants import MIN_SUPPORTED_FILE_VERSION
from .constants import NOT_APPLICABLE_H5_METADATA
from .constants import ORIGINAL_FILE_VERSION_UUID
from .constants import PLATE_BARCODE_UUID
from .constants import REF_SAMPLING_PERIOD_UUID
from .constants import REFERENCE_SENSOR_READINGS
from .constants import REFERENCE_VOLTAGE_UUID
from .constants import SLEEP_FIRMWARE_VERSION_UUID
from .constants import SOFTWARE_BUILD_NUMBER_UUID
from .constants import SOFTWARE_RELEASE_VERSION_UUID
from .constants import START_RECORDING_TIME_INDEX_UUID
from .constants import TISSUE_SAMPLING_PERIOD_UUID
from .constants import TISSUE_SENSOR_READINGS
from .constants import TOTAL_WELL_COUNT_UUID
from .constants import TRIMMED_TIME_FROM_ORIGINAL_END_UUID
from .constants import TRIMMED_TIME_FROM_ORIGINAL_START_UUID
from .constants import USER_ACCOUNT_ID_UUID
from .constants import UTC_BEGINNING_DATA_ACQUISTION_UUID
from .constants import UTC_BEGINNING_RECORDING_UUID
from .constants import UTC_FIRST_REF_DATA_POINT_UUID
from .constants import UTC_FIRST_TISSUE_DATA_POINT_UUID
from .constants import UTC_TIMESTAMP_OF_FILE_VERSION_MIGRATION_UUID
from .constants import WELL_COLUMN_UUID
from .constants import WELL_INDEX_UUID
from .constants import WELL_NAME_UUID
from .constants import WELL_ROW_UUID
from .constants import XEM_SERIAL_NUMBER_UUID
from .exceptions import FileAttributeNotFoundError
from .exceptions import MantarrayFileNotLatestVersionError
from .exceptions import UnsupportedFileMigrationPath
from .exceptions import UnsupportedMantarrayFileVersionError
from .exceptions import WellRecordingsNotFromSameSessionError
from .file_writer import MantarrayH5FileCreator
from .file_writer import migrate_to_latest_version
from .file_writer import migrate_to_next_version
from .files import BasicWellFile
from .files import PlateRecording
from .files import WELL_FILE_CLASSES
from .files import WellFile
from .files import WellFile_0_3_1
from .files import WellFile_0_4_1
from .files import WellFile_0_4_2


__all__ = [
    "WellFile",
    "PlateRecording",
    "UTC_BEGINNING_DATA_ACQUISTION_UUID",
    "FILE_FORMAT_VERSION_METADATA_KEY",
    "START_RECORDING_TIME_INDEX_UUID",
    "CUSTOMER_ACCOUNT_ID_UUID",
    "USER_ACCOUNT_ID_UUID",
    "SOFTWARE_BUILD_NUMBER_UUID",
    "SOFTWARE_RELEASE_VERSION_UUID",
    "MAIN_FIRMWARE_VERSION_UUID",
    "SLEEP_FIRMWARE_VERSION_UUID",
    "XEM_SERIAL_NUMBER_UUID",
    "MANTARRAY_NICKNAME_UUID",
    "REFERENCE_VOLTAGE_UUID",
    "WELL_NAME_UUID",
    "WELL_ROW_UUID",
    "WELL_COLUMN_UUID",
    "WELL_INDEX_UUID",
    "TOTAL_WELL_COUNT_UUID",
    "REF_SAMPLING_PERIOD_UUID",
    "TISSUE_SAMPLING_PERIOD_UUID",
    "ADC_GAIN_SETTING_UUID",
    "PLATE_BARCODE_UUID",
    "ADC_TISSUE_OFFSET_UUID",
    "ADC_REF_OFFSET_UUID",
    "MANTARRAY_SERIAL_NUMBER_UUID",
    "UTC_BEGINNING_RECORDING_UUID",
    "UTC_FIRST_TISSUE_DATA_POINT_UUID",
    "UTC_FIRST_REF_DATA_POINT_UUID",
    "HARDWARE_TEST_RECORDING_UUID",
    "CURI_BIO_ACCOUNT_UUID",
    "CURI_BIO_USER_ACCOUNT_ID",
    "METADATA_UUID_DESCRIPTIONS",
    "DATETIME_STR_FORMAT",
    "CENTIMILLISECONDS_PER_SECOND",
    "MICROSECONDS_PER_CENTIMILLISECOND",
    "WellRecordingsNotFromSameSessionError",
    "MIN_SUPPORTED_FILE_VERSION",
    "UnsupportedMantarrayFileVersionError",
    "FileAttributeNotFoundError",
    "BACKEND_LOG_UUID",
    "COMPUTER_NAME_HASH_UUID",
    "BARCODE_IS_FROM_SCANNER_UUID",
    "IS_FILE_ORIGINAL_UNTRIMMED_UUID",
    "TRIMMED_TIME_FROM_ORIGINAL_START_UUID",
    "TRIMMED_TIME_FROM_ORIGINAL_END_UUID",
    "ORIGINAL_FILE_VERSION_UUID",
    "CURRENT_HDF5_FILE_FORMAT_VERSION",
    "MantarrayH5FileCreator",
    "FILE_MIGRATION_PATHS",
    "migrate_to_next_version",
    "migrate_to_latest_version",
    "UnsupportedFileMigrationPath",
    "BasicWellFile",
    "WellFile_0_4_1",
    "WellFile_0_3_1",
    "WellFile_0_4_2",
    "file_writer",
    "WELL_FILE_CLASSES",
    "UTC_TIMESTAMP_OF_FILE_VERSION_MIGRATION_UUID",
    "FILE_VERSION_PRIOR_TO_MIGRATION_UUID",
    "NOT_APPLICABLE_H5_METADATA",
    "MantarrayFileNotLatestVersionError",
    "TISSUE_SENSOR_READINGS",
    "REFERENCE_SENSOR_READINGS",
]



# -*- coding: utf-8 -*-
"""Setup configuration."""
from setuptools import find_packages
from setuptools import setup


setup(
    name="mantarray_file_manager",
    version="0.4.6",
    description="Finds and opens Mantarray files.",
    url="https://github.com/CuriBio/mantarray-file-manager",
    author="Curi Bio",
    author_email="contact@curibio.com",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=[
        "h5py>=2.10.0",
        "nptyping>=1.3.0",
        "numpy>=1.19.1",
        "stdlib-utils>=0.2.1",
        "semver>=2.10.2",
        "immutable_data_validation>=0.2.1",
        "immutabledict>=1.1.0",
    ],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
    ],
)
